@@ -1453,40 +1453,44 @@ shttp_print_response_status (struct shttp_connection *conn)
   if (!dump && conn->result && conn->result->type == json_arr)
     {
       size_t i, n = json_array_size (conn->result);
-      
-      for (i = 0; i < n; i++)
+
+      if (n == 0)
 	{
-	  struct json_value *elt, *v;
-	  
-	  if (json_array_get (conn->result, i, &elt) == 0)
-	    {
-	      if (elt->type != json_object)
-		{
-		  grecs_error (NULL, 0, _("%lu: unexpected value type"),
-			       (unsigned long) i);
-		  print_json (stderr, v);
-		  continue;
-		}
-	      if (json_object_get_type (elt, "tag", json_string, &v) == 0)
-		{
-		  printf ("%s: ", v->v.s);
-		  if (json_object_get_type (elt, "status", json_string, &v)
-		      == 0)
-		    {
-		      if (strcmp (v->v.s, "OK") == 0)
-			fputs (v->v.s, stdout);
-		      else if (json_object_get_type (elt, "message",
-						     json_string, &v) == 0)
-			fputs (v->v.s, stdout);
-		      else
-			printf (_("unknown error"));
-		    }
-		  else
-		    printf (_("unknown status"));
-		  fputc ('\n', stdout);
-		}
-	    }
+	  grecs_error (NULL, 0, _("no component matches the supplied condition"));
 	}
+      else
+	for (i = 0; i < n; i++)
+	  {
+	    struct json_value *elt, *v;
+
+	    if (json_array_get (conn->result, i, &elt) == 0)
+	      {
+		if (elt->type != json_object)
+		  {
+		    grecs_error (NULL, 0, _("%lu: unexpected value type"),
+				 (unsigned long) i);
+		    print_json (stderr, v);
+		    continue;
+		  }
+		if (json_object_get_type (elt, "tag", json_string, &v) == 0)
+		  {
+		    printf ("%s: ", v->v.s);
+		    if (json_object_get_type (elt, "status", json_string, &v) == 0)
+		      {
+			if (strcmp (v->v.s, "OK") == 0)
+			  fputs (v->v.s, stdout);
+			else if (json_object_get_type (elt, "message",
+						       json_string, &v) == 0)
+			  fputs (v->v.s, stdout);
+			else
+			  printf (_("unknown error"));
+		      }
+		    else
+		      printf (_("unknown status"));
+		    fputc ('\n', stdout);
+		  }
+	      }
+	  }
     }
 }
 
